/*******************************************************************************
*                         Goggles Music Manager                                *
********************************************************************************
*           Copyright (C) 2007-2009 by Sander Jansen. All Rights Reserved      *
*                               ---                                            *
* This program is free software: you can redistribute it and/or modify         *
* it under the terms of the GNU General Public License as published by         *
* the Free Software Foundation, either version 3 of the License, or            *
* (at your option) any later version.                                          *
*                                                                              *
* This program is distributed in the hope that it will be useful,              *
* but WITHOUT ANY WARRANTY; without even the implied warranty of               *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
* GNU General Public License for more details.                                 *
*                                                                              *
* You should have received a copy of the GNU General Public License            *
* along with this program.  If not, see http://www.gnu.org/licenses.           *
********************************************************************************/
#include <xincs.h>

// Needed for pthread_kill
#if defined(__FreeBSD__) || defined(__OpenBSD__)
#include <pthread.h>
#endif
#include <signal.h>

#include "common.h"
#include "GMThread.h"
#include "GMFetch.h"
//#include "download.h"
#include "GMList.h"
#include "GMSource.h"
#include "GMPlayerManager.h"
#include "GMURL.h"

#include <FXDLL.h>

typedef struct url* (*func_downloadParseURL)(const char *);
typedef FILE* (*func_downloadGet) (struct url *, const char *);
typedef void (*func_downloadFreeURL)(struct url *);
#if FOXVERSION < FXVERSION(1,7,0)
static void * dll_download = NULL;
#else
static FXDLL dll_download;
#endif
static int dll_download_ref = 0;

static func_downloadGet             dll_downloadGet          = NULL;
static func_downloadParseURL        dll_downloadParseURL     = NULL;
static func_downloadFreeURL         dll_downloadFreeURL      = NULL;
static int *                        dll_downloadRestartCalls = NULL;


GMFetch::GMFetch(FXObject * tgt) : GMThread(tgt) {
#if FOXVERSION < FXVERSION(1,7,0)
  if (dll_download==NULL) {
    dll_download = fxdllOpen("libdownload.so");
    if (dll_download) {
      dll_downloadGet          = (func_downloadGet)      fxdllSymbol(dll_download,"downloadGet");
      dll_downloadParseURL     = (func_downloadParseURL) fxdllSymbol(dll_download,"downloadParseURL");
      dll_downloadFreeURL      = (func_downloadFreeURL)  fxdllSymbol(dll_download,"downloadFreeURL");
      dll_downloadRestartCalls = (int*)                  fxdllSymbol(dll_download,"downloadRestartCalls");
      }
    else {
      dll_download = fxdllOpen("libfetch.so");
      if (dll_download) {
        dll_downloadGet          = (func_downloadGet)      fxdllSymbol(dll_download,"fetchGet");
        dll_downloadParseURL     = (func_downloadParseURL) fxdllSymbol(dll_download,"fetchParseURL");
        dll_downloadFreeURL      = (func_downloadFreeURL)  fxdllSymbol(dll_download,"fetchFreeURL");
        dll_downloadRestartCalls = (int*)                  fxdllSymbol(dll_download,"fetchRestartCalls");
        }
      }
    }
  if (dll_download) dll_download_ref++;
#else
  if (!dll_download.loaded()) {
    if (dll_download.load("libdownload.so")) {
      dll_downloadGet          = (func_downloadGet)      dll_download.address("downloadGet");
      dll_downloadParseURL     = (func_downloadParseURL) dll_download.address("downloadParseURL");
      dll_downloadFreeURL      = (func_downloadFreeURL)  dll_download.address("downloadFreeURL");
      dll_downloadRestartCalls = (int*)                  dll_download.address("downloadRestartCalls");
      }
    else if (dll_download.load("libfetch.so")) {
      dll_downloadGet          = (func_downloadGet)      dll_download.address("fetchGet");
      dll_downloadParseURL     = (func_downloadParseURL) dll_download.address("fetchParseURL");
      dll_downloadFreeURL      = (func_downloadFreeURL)  dll_download.address("fetchFreeURL");
      dll_downloadRestartCalls = (int*)                  dll_download.address("fetchRestartCalls");
      }
    }
  if (dll_download.loaded()) dll_download_ref++;
#endif
  }

FXbool getLine(FILE * fp,FXString & line){
  FXint c;
  while((c=fgetc(fp))!=EOF){
    if (c=='\r') continue;
    if (c=='\n') return true;
    line+=(FXchar)c;
    }
  return false;
  }


void GMFetch::parseM3U(FILE * fp){
  FXString line;
  while(getLine(fp,line) && running){
    if (line[0]=='#') continue;
    mrl.append(line);
    line.clear();
    }
  }

void GMFetch::parsePLS(FILE * fp){
  FXString line;
  while(getLine(fp,line) && running){
    if (compare(line,"File",4)==0) {
      mrl.append(line.after('='));
      }
    line.clear();
    }
  }


FXint GMFetch::run() {
  struct url * uri = dll_downloadParseURL(filename.text());
  if (uri==NULL) {
    feedback.message(target,FXSEL(SEL_COMMAND,GMPlayerManager::ID_DOWNLOAD_COMPLETE),NULL,0);
    return 0;
    }
  (*dll_downloadRestartCalls)=0;
  FILE * fp  = dll_downloadGet(uri,"r");
  if (fp) {
    if (comparecase(FXPath::extension(GMURL::path(filename)),"m3u")==0){
      parseM3U(fp);
      }
    else if (comparecase(FXPath::extension(GMURL::path(filename)),"pls")==0){
      parsePLS(fp);
      }
    fclose(fp);
    }
  dll_downloadFreeURL(uri);
  feedback.message(target,FXSEL(SEL_COMMAND,GMPlayerManager::ID_DOWNLOAD_COMPLETE),NULL,0);
  return 0;
  }


void GMFetch::get(const FXString & file) {
  filename=file;
#if FOXVERSION < FXVERSION(1,7,0)
  if (dll_download==NULL)
#else
  if (!dll_download.loaded())
#endif
    feedback.message(target,FXSEL(SEL_COMMAND,GMPlayerManager::ID_DOWNLOAD_COMPLETE),NULL,0);
  else
    start();
  }

void GMFetch::dispose() {
  GMThread::dispose();
  if (id() && FXThread::running() ) {
    pthread_kill((pthread_t)id(),SIGINT);
    }
  }

GMFetch::~GMFetch(){
#if FOXVERSION < FXVERSION(1,7,0)
  if (dll_download) {
    dll_download_ref--;
    if (dll_download_ref==0) {
      fxdllClose(dll_download);
      dll_download=NULL;
      }
    }
#else
  if (dll_download.loaded()){
    dll_download_ref--;
    if (dll_download_ref==0) {
      dll_download.unload();
      }
    }
#endif
  }
